# Copyright 2016 Pavle Jonoski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from threading import Condition


class Promise:
    def __init__(self):
        self.error = None
        self.__result = None
        self.__is_done = False
        self.__cw = Condition()

    @property
    def result(self):
        with self.__cw:
            while not self.__is_done:
                self.__cw.wait()

        if self.error:
            raise DistributedException(self.error, self.__result)
        return self.__result

    @property
    def is_done(self):
        return self.__is_done

    def complete(self, result=None, error=None):
        self.__is_done = True
        self.error = error
        self.__result = result
        with self.__cw:
            self.__cw.notify_all()



class DistributedException(Exception):

    def __init__(self, message, data=None):
        super(DistributedException, self).__init__(message)
        self.data = data



import sys

sys.path.append('..')

print(sys.path)

from troup.infrastructure import AsyncIOWebSocketServer, IncomingChannelWSAdapter
from ws4py.async_websocket import EchoWebSocket

server = AsyncIOWebSocketServer(host='', port=8910, web_socket_class=IncomingChannelWSAdapter)

def run_server():
    print('Srv run')
    server.start()
    print('Srv end')

import threading

t = threading.Thread(target=run_server)

t.start()



import sys

sys.path.append('..')

from troup.process import SSHRemoteProcess

rp = SSHRemoteProcess(id="xterm-01", name="xterm", args=[], cwd="/home/pavle", target_host="natemago", target_port="7799", ssh_user="pavle", forward_video=True)

import time

rp.execute()

time.sleep(10)



# Copyright 2016 Pavle Jonoski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = 'pavle'

from types import FunctionType
from types import MethodType
from troup.observer import Observable
import logging
from queue import Queue


class ChannelError(Exception):
    pass


class ChannelClosedError(ChannelError):
    pass


class Channel:

    CREATED = 'CREATED'
    CONNECTING = 'CONNECTING'
    OPEN = 'OPEN'
    CLOSING = 'CLOSING'
    CLOSED = 'CLOSED'
    ERROR = 'ERROR'

    def __init__(self, name, to_url):
        self.name = name
        self.status = 'CREATED'
        self.listeners = []
        self.event_listeners = {}
        self.to_url = to_url
        self.log = logging.getLogger(self.__class__.__name__)

    def open(self):
        if self.status is not Channel.CREATED:
            raise ChannelError('Unable to open channel')
        try:
            self.status = Channel.CONNECTING
            self.connect()
            self.status = Channel.OPEN
        except ChannelError:
            self.status = Channel.ERROR
            raise
        except Exception as e:
            self.status = Channel.ERROR
            raise ChannelError() from e

    def close(self):
        if self.status is not Channel.OPEN:
            raise ChannelError('Unable to close channel')
        try:
            self.status = Channel.CLOSING
            self.disconnect()
            self.status = Channel.CLOSED
        except ChannelError:
            self.status = Channel.ERROR
            raise
        except Exception as e:
            self.status = Channel.ERROR
            raise ChannelError() from e

    def connect(self):
        pass

    def disconnect(self):
        pass

    def register_listener(self, callback):
        listener = self.__wrap_listener__(callback)
        self.listeners.append(listener)

    def __wrap_listener__(self, callback):
        return ListenerWrapper(callback)

    def send(self, data):
        self.log.debug('[CH<Channel>: %s]: empty send' % self.name)

    def data_received(self, data):
        for listener in self.listeners:
            try:
                listener.on_data(data)
            except Exception as e:
                self.log.exception('Listener error: %s', e)

    def on(self, event_name, callback):
        callbacks = self.event_listeners.get(event_name)
        if not callbacks:
            callbacks = self.event_listeners[event_name] = []
        if not callback in callbacks:
            callbacks.append(callback)

    def trigger(self, event, *data):
        callbacks = self.event_listeners.get(event)
        if callbacks:
            for callback in callbacks:
                try:
                    callback(*data)
                except Exception as e:
                    self.log.debug('An error while triggering event {}', event, e)

    def __repr__(self):
        return '<Channel %s> to %s' % (self.name, self.to_url)


class ListenerWrapper:
    def __init__(self, delegate):
        self.delegate = self.__get_callable__(delegate)

    def __get_callable__(self, delegate):
        if isinstance(delegate, FunctionType) or \
           isinstance(delegate, MethodType):
            return delegate
        else:
            if hasattr(delegate, 'on_data'):
                return getattr(delegate, 'on_data')
        raise ChannelError('Invalid listener. It is not a callable object and does not contain on_data method.')

    def on_data(self, data):
        self.delegate(data)


from ws4py.async_websocket import WebSocket
from threading import Event


class IncommingChannel(Channel):

    def __init__(self, name, to_url, adapter=None):
        super(IncommingChannel, self).__init__(name, to_url)
        self.adapter = adapter
        self.close_event = Event()

    def disconnect(self):
        self.adapter.close(code=1000, reason="client-closing")
        # TODO: Wait to actually close
        self.close_event.wait()

    def notify_close(self):
        self.close_event.set()

    def send(self, data):
        if self.status is Channel.OPEN:
            self.adapter.send(payload=data)
        else:
            raise ChannelError('Not open')


class IncomingChannelWSAdapter(WebSocket):

    def __init__(self, protocol):
        WebSocket.__init__(self, protocol)
        self.server = None
        self.log = logging.getLogger(self.__class__.__name__)
        self.channel = None

    def opened(self):
        self.server = self.proto.server
        try:
            self.channel = IncommingChannel(
                name="channel[%s-%s]" % (self.local_address, self.peer_address),
                to_url=str(self.peer_address),
                adapter=self)
            self.channel.open()
            self.server.on_channel_open(self.channel)
        except Exception as e:
            logging.exception(e)
            raise e

    def closed(self, code, reason=None):
        self.log.debug('closing ws. code=%s, reason=%s'%(str(code),str(reason)))
        self.channel.notify_close()
        self.server.on_channel_closed(self.channel)

    def received_message(self, message):
        #print(' -> %s' % str(message))
        #print('Message is text %s - data[%s]' % (message.is_text,message.data))
        try:
            self.channel.data_received(str(message))
        except Exception as e:
            logging.exception(e)

    @property
    def local_address(self):
        """
        Local endpoint address as a tuple
        """
        if not self._local_address:
            self._local_address = self.proto.reader._transport.get_extra_info('sockname')
            if len(self._local_address) == 4:
                self._local_address = self._local_address[:2]
        return self._local_address

    @property
    def peer_address(self):
        """
        Peer endpoint address as a tuple
        """
        if not self._peer_address:
            self._peer_address = self.proto.reader._transport.get_extra_info('peername')
            if len(self._peer_address) == 4:
                self._peer_address = self._peer_address[:2]
        return self._peer_address


from ws4py.server.tulipserver import WebSocketProtocol


class ServerAwareWebSocketProtocol (WebSocketProtocol):

    def __init__(self, handler_class, server):
        super(ServerAwareWebSocketProtocol, self).__init__(handler_class)
        self.server = server


import asyncio


class AsyncIOWebSocketServer:

    def __init__(self, host='', port=1700, web_socket_class=IncomingChannelWSAdapter):
        self.host = host
        self.port = port
        self.web_socket_class = web_socket_class
        self.aio_loop = asyncio.get_event_loop()
        self.running = False
        self.channels = {}
        self.listeners = []
        self.aio_sf = None
        self.server_address = None
        self.log = logging.getLogger('AsyncIOWebSocketServer')

    def start(self):
        proto = lambda: ServerAwareWebSocketProtocol(self.web_socket_class, self)
        asyncio.set_event_loop(self.aio_loop)
        sf = self.aio_loop.create_server(proto, self.host, self.port)
        s = self.aio_loop.run_until_complete(sf)
        self.server_address = s.sockets[0].getsockname()
        self.log.info('Server stared on %s' % str(s.sockets[0].getsockname()))
        self.aio_sf = sf
        self.aio_loop.run_forever()
        self.aio_loop.close()
        self.log.debug('Async Event loop closed.')


    def stop(self):
        def stop_server_and_loop():
            self.aio_sf.close()
            self.aio_loop.stop()
            self.log.debug('Server closed. Event loop notified for stop.')
        self.aio_loop.call_soon_threadsafe(stop_server_and_loop)

    def on_channel_open(self, channel):
        self.channels[channel.name] = channel
        self.log.debug('Channel %s => %s added' % (channel.name, channel))
        self.notify_event('channel.open', channel)

    def on_channel_closed(self, channel):
        del self.channels[channel]
        self.notify_event('channel.closed', channel)

    def on_event(self, callback):
        self.listeners.append(callback)

    def notify_event(self, event, channel):
        for listener in self.listeners:
            listener(event, channel)

    def get_server_endpoint(self):
        return 'ws://%s:%s' % (self.host or 'localhost', self.port)
# -- outgoing connection

from ws4py.client.threadedclient import WebSocketClient


class OutgoingChannelWSAdapter(WebSocketClient):

    def __init__(self, url, handlers):
        super(OutgoingChannelWSAdapter, self).__init__(url=url)
        self.handlers = handlers

    def __noop__(self, *args, **kwargs):
        pass

    def __handler__(self, name):
        return self.handlers.get(name) or self.__noop__

    def opened(self):
        self.__handler__('opened')()

    def closed(self, code, reason=None):
        self.__handler__('closed')(code, reason)

    def received_message(self, m):
        if m and m.data:
            if m.is_text:
                self.__handler__('on_data')(str(m))
            else:
                self.__handler__('on_data')(m.data)


class OutgoingChannelOverWS(Channel):
    
    def __init__(self, name, to_url, early_messages='queue', queue_max_size=1000):
        super(OutgoingChannelOverWS, self).__init__(name, to_url)
        self.web_socket = OutgoingChannelWSAdapter(url=to_url,
                                                   handlers={
                                                       'opened': self._on_open_handler_,
                                                       'closed': self._on_closed_handler_,
                                                       'on_data': self.data_received
                                                   })
        self._early_messages = early_messages
        self._queue_max_size = queue_max_size
        self.queue = None
        self.__setup_early_strategy()
        
    def __setup_early_strategy(self):
        if self._early_messages == 'queue':
            self.queue = Queue(maxsize=self._queue_max_size)
    
    def __handle_early_messages(self):
        if self._early_messages == 'queue':
            while not self.queue.empty():
                msg = self.queue.get_nowait()
                self.send(msg)
                
    
    def _on_open_handler_(self):
        self.trigger('open', self)
        self.__handle_early_messages()
        self.on_opened()

    def on_opened(self):
        pass

    def _on_closed_handler_(self, code, reason=None):
        self.trigger('closed', self, code, reason)
        self.status = Channel.CLOSING
        self.on_closed(code, reason)
        self.status = Channel.CLOSED

    def on_closed(self, code, reason=None):
        pass

    def connect(self):
        try:
            self.web_socket.connect()
        except (ConnectionRefusedError, ConnectionAbortedError, ConnectionResetError) as e:
            raise ChannelClosedError() from e

    def disconnect(self):
        self.web_socket.close()

    def send(self, data):
        if self.status == Channel.OPEN:
            try:
                self.web_socket.send(payload=data)
            except (ConnectionRefusedError, ConnectionAbortedError, ConnectionResetError) as e:
                raise ChannelClosedError() from e
        elif self.status in [Channel.CREATED, Channel.CONNECTING]:
            self.__send_early(data)
        else:
            raise ChannelClosedError('Cannot send: invalid channel status')
    
    def __send_early(self, data):
        if self._early_messages == 'queue':
            self.queue.put(data)
        elif self._early_messages == 'reject':
            raise Exception('Early message rejected')
        else:
            logging.warn('Early message [%s] not send due to unknown early messages strategy: %s' %
                         (str(data), self._early_messages))
            

class ChannelManager(Observable):

    def __init__(self, aio_server):
        #self.config = config
        super(ChannelManager, self).__init__()
        self.aio_server = aio_server
        self.channels = {}
        self.by_url = {}
        self.log = logging.getLogger('channel-manager')
        self.aio_server.on_event(self._aio_server_event_)

    def _aio_server_event_(self, event, channel):
        if event == 'channel.open':
            self._on_open_channel_(channel)
        elif event == 'channel.closed':
            pass
        else:
            pass

    def channel(self, name=None, to_url=None):
        if self.channels.get(name):
            return self.channels[name]
        if to_url and self.by_url.get(to_url):
            return self.by_url[to_url]
        if not to_url:
            raise Exception('No channel URL specified')
        if not name and to_url:
            name = to_url
        channel = self.open_channel_to(name, to_url)
        self.channels[name] = channel
        self.by_url[to_url] = channel
        return channel

    def open_channel_to(self, name, url):
        och = OutgoingChannelOverWS(name=name, to_url=url)
        self._on_open_channel_(och)
        try:
            och.open()
        except ChannelClosedError:
            self.trigger('channel.closed', och)
            raise
        return och

    def close_channel(self, name=None, endpoint=None):
        pass

    def _on_open_channel_(self, channel):
        channel.on('closed', self._handle_closed_channel_)

        def get_data_listener(chn):
            def data_listener(data):
                self.trigger('channel.data', data, chn)
            return data_listener

        channel.register_listener(get_data_listener(channel))

        self.trigger('channel.open', channel)

    def _handle_closed_channel_(self, channel, code, reason=None):
        del self.channels[channel.name]
        del self.by_url[channel.to_url]
        self.trigger('channel.closed', channel)

    def listen(self, name=None, to_url=None, listener=None):
        channel = self.channel(name, to_url)
        channel.register_listener(listener)

    def send(self, name=None, to_url=None, data=None):
        channel = self.channel(name, to_url)
        try:
            channel.send(data)
        except ChannelClosedError as e:
            channel.close()
            self._handle_closed_channel_(channel, 1006, str(e))

    def on_data(self, callback, from_channel=None):
        def actual_callback_no_filter(data, chn):
            callback(data)

        def actual_callback_with_filter(data, channel):
            if channel.name == from_channel:
                callback(data)

        if from_channel:
            self.on('channel.data', actual_callback_with_filter)
        else:
            self.on('channel.data', actual_callback_no_filter)


# -- simplest message bus in the world
class MessageHandler:

    def __init__(self, handler, message_filter):
        self.handler = handler
        self.message_filter = message_filter

    def __call__(self, message):
        if self.message_filter:
            if not self.message_filter(message):
                return
        self.handler(message)

    def __eq__(self, other):
        if not type(self) is type(other):
            return False
        if self.handler and other.handler:
            if not self.handler.__eq__(other.handler):
                return False
            if self.message_filter is not None:
                if not other.message_filter:
                    return False
                return self.message_filter.__eq__(other.message_filter)
            else:
                return not other.message_filter

    def __hash__(self):
        return self.handler.__hash__()


class MessageBus:

    def __init__(self):
        self.subscribers = {}
        self.log = logging.getLogger(self.__class__.__name__)

    def on(self, topic, handler, message_filter=None):
        if not handler:
            raise Exception('Handler not specified')
        if not topic:
            raise Exception('Topic not specified')
        subscribers = self.__get_subscribers__(topic)
        if handler in subscribers:
            raise Exception('Handler already registered')
        self.log.debug('Listening on topic %s. Handler %s (filter=%s)', topic, handler, message_filter)
        subscribers.append(handler)

    def __get_subscribers__(self, topic):
        subscribers = self.subscribers.get(topic)
        if not subscribers:
            subscribers = []
            self.subscribers[topic] = subscribers
        return subscribers

    def publish(self, topic, *events):
        subscribers = self.subscribers.get(topic)
        if subscribers:
            for handler in subscribers:
                try:
                    handler(*events)
                except Exception as e:
                    self.log.exception(e)

    def remove(self, topic, handler):
        subscribers = self.subscribers.get(topic)
        if subscribers:
            subscribers.remove(handler)


message_bus = MessageBus()


class Subscribe:
    def __init__(self, topic, filter=None, bus=None):
        self.topic = topic
        self.filter = filter
        self.bus = bus
        if not self.bus:
            self.bus = message_bus

    def __call__(self, method):
        self.bus.on(self.topic, method)
        return method


class Bus:

    def __init__(self):
        self.subscribe = Subscribe

bus = Bus()



import random

class TestSyncManager:
    
    def __init__(self):
        self.nodes = ['a','b','c','d','e','f']
        self.random_buffer = []
        self.shuffle_all()
    
    def shuffle_all(self):
        nodes = [] + self.nodes
        random.shuffle(nodes)
        self.random_buffer = self.random_buffer + nodes
        return self.random_buffer
    
    def next(self, n):
        while len(self.random_buffer) < n:
            self.shuffle_all()
        shuffled = self.random_buffer[0:n]
        self.random_buffer = self.random_buffer[n:]
        return shuffled
                
tm = TestSyncManager()

for i in range(1, 10):
    #print('--------------')
    #print('before: ' + str(tm.random_buffer))
    print(tm.next(40))
    #print('after:  ' + str(tm.random_buffer))
    



__author__ = 'pavle'

import sys

sys.path.append('..')

from troup.client import CommandAPI, client_to_local_node, ChannelClient
import time

cc = ChannelClient(nodes_specs=['RPI:ws://192.168.2.128:7000'])

cmd = CommandAPI(channel_client=cc)

promise = cmd.send(CommandAPI.task('LocalProcess', {
    'directory': '/home/piadmin',
    'executable': '/bin/ls',
    'args': ['-l', '-a']
}, track_out=True, ttl=10000))
print(promise)
print('cmd send')
try:
    result = promise.result
    print('Result -> %s' % str(result))
    time.sleep(2)
    def on_task_result(res):
        print('Task result %s' % res)

    task_result_cmd = CommandAPI.command('task-result', {'task-id': result})
    promise = cmd.send(task_result_cmd, on_reply=on_task_result)
    print('Task result request send')
    ls_res = promise.result
    print('Result: %s' % ls_res)
finally:
    cmd.shutdown()
    print('done')



import asyncio

import threading

class TestProtocol(asyncio.Protocol):
    
    def connection_made(self, transport):
        print('connection_made')
    
    def data_received(self, data):
        print('data_received')
    
    def connection_lost(self, exc):
        print('connection_lost')    
     

loop = asyncio.get_event_loop()
server = loop.run_until_complete(loop.create_server(TestProtocol, 'localhost', 15001))

def run_server():
    print('run_forever...')
    loop.run_forever()
    print('server exit')
    loop.close()
    print('loop closed')

#@asyncio.coroutine
def stop_loop(loop):
    print('loop stop')
    loop.stop()
    print('loop indeed stop')
    

srv_thread = threading.Thread(target=run_server)

srv_thread.start()

import time

time.sleep(1)

def stop_server():
    server.close()
    print('server close')
    loop.call_soon_threadsafe(stop_loop, loop) # must schedule with threadsafe


stop_server()

print('all out')



# Copyright 2016 Pavle Jonoski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = 'pavle'

from troup.store import InMemorySyncedStore
from troup.infrastructure import AsyncIOWebSocketServer, IncomingChannelWSAdapter, ChannelManager, message_bus, bus
from troup.system import StatsTracker, SystemStats
from troup.messaging import message, serialize, deserialize, deserialize_dict, Message
import threading
from troup.threading import IntervalTimer
from troup.apps import App
from troup.process import this_process_info_file, open_process_lock_file
from troup.tasks import TasksRunner, build_task, task_for_app
import random
from math import ceil
from os import getpid
from functools import reduce
import logging


class Node:

    def __init__(self, node_id, config, store=None, channel_manager=None,
                 aio_server=None, stats_tracker=None, sync_manager=None,
                 tasks_runner=None):
        self.node_id = node_id

        self.log = logging.getLogger('Node(%s)' % self.node_id)

        self.config = config
        self.store = store or self._build_store_()
        self.channel_manager = channel_manager or None
        self.aio_server = aio_server or None
        self.stats_tracker = stats_tracker or None
        self.sync_manager = sync_manager or None
        self.bus = message_bus
        self.lock = None
        self.pid = getpid()
        self.commands = {}
        self.runner = tasks_runner or TasksRunner(max_workers=int(self.config.get('runner-max-workers', '3')))

        self.__register_commands()

    def __lock(self):
        if self.lock:
            return self.lock
        self.__check_lock()
        self.lock = local_node_lock_file({"name": self.node_id})
        return self.lock

    def __check_lock(self):
        lock = check_local_node_lock()
        if lock and lock.pid != self.pid:
            raise Exception(('Seems there is a node already running. PID is %d. ' +
                            'If no node with this pid is running, then you may ' +
                            'have to delete the file manually.') % lock.pid)

    def _start_channel_manager_(self):
        if self.channel_manager is not None:
            return self.channel_manager
        aio_srv = self.aio_server or AsyncIOWebSocketServer(host=self.config['server'].get('hostname'),
                                         port=self.config['server']['port'],
                                         web_socket_class=IncomingChannelWSAdapter)

        def start_aio_server():
            self.log.debug('AIO Server start')
            aio_srv.start()
            self.log.debug('AIO Server end')

        th = threading.Thread(target=start_aio_server)
        th.start()
        channel_manager = ChannelManager(aio_srv)
        self.aio_server = aio_srv
        self.log.debug('AIO Server set up')
        return channel_manager

    def __register_message_dispatcher__(self):
        def on_channel_data(message_str, channel):
            try:
                msg = deserialize(message_str)
                if msg.headers.get('type'):
                    self.bus.publish(msg.headers['type'], msg, channel)
                else:
                    self.bus.publish('__message.genericType', msg, channel)
            except Exception as e:
                self.log.exception('Failed to handle channel data', e)
                self.lock.debug('Message ', message_str)
        self.channel_manager.on('channel.data', on_channel_data)

    def _build_store_(self):
        store = InMemorySyncedStore(root_path=self.config['store']['path'])
        return store

    def _start_stats_tracker_(self):
        self.stats_tracker = self.stats_tracker or StatsTracker(period=self.config['stats']['update_interval'])
        self.log.info('stats tracking ON')

    def _start_sync_manager_(self):
        self.sync_manager = self.sync_manager or SyncManager(node=self, channel_manager=self.channel_manager,
                                        event_processor=None, sync_interval=10000, sync_percent=0.3)
        self.sync_manager.start()

        neighbours = self.config.get('neighbours')
        if neighbours:
            for url in neighbours:
                name, sep, endpoint = url.partition(':')
                node = NodeInfo(name=name, stats=None, apps=None, endpoint=endpoint)
                self.sync_manager.register_node(node)
                self.log.debug('Added neighbour %s [%s]' % (name, endpoint))

    def __register_to_local(self):
        if self.config.get('lock'):
            endpoint = self.aio_server.get_server_endpoint()
            self.lock.set_info('url', endpoint)

    def __register_command_handlers(self):
        @bus.subscribe('task')
        def __on_task__(task, inc_channel):
            try:
                run = self.runner.run(build_task(task))
                # FIXME: Add context to runner.
                self.__reply(task, run.id, inc_channel)
            except Exception as e:
                self.log.exception('Failed to run task %s', task)
                self.__reply(task, str(e), inc_channel, error=True)

        @bus.subscribe('command')
        def __on_command__(command, inc_channel):
            self.log.debug('Received command: %s over channel %s' % (command, inc_channel))
            self.__process_command(command, inc_channel)

    def __register_commands(self):
        self.command_handler('apps', self.__list_apps)
        self.command_handler('info', self.__get_info)
        self.command_handler('task-result', self.__task_result)
        self.command_handler('run-app', self.__run_app)

    def __run_app(self, command):
        print('RUN APP COMMAND RECEIVED: %s' % command)
        print("RUN APP %s" % command.data['app'])
        return self.run_app(app_name=command.data['app'])

    def __list_apps(self, command):
        return self.get_available_apps()

    def __get_info(self, command):
        return self.get_node_info()

    def __task_result(self, command):
        stats = self.runner.stats
        task_id = command.data['task-id']
        status = stats.get(task_id)
        if not status:
            raise Exception('No such task')
        if status['status'] is not 'DONE':
            raise Exception('Task status %s' % status['status'])
        run = self.runner.tasks[task_id]
        self.log.debug('Task result: [%s]' % run.task.result)
        return run.task.result

    def command_handler(self, command, handler):
        self.commands[command] = handler

    def __process_command(self, command, channel):
        handler = self.commands.get(command.headers['command'])
        if not handler:
            self.__reply(command, reply='Unknown command', error=True, channel=channel)
            return
        try:
            reply = handler(command)
            self.__reply(command, reply=reply, channel=channel)
        except Exception as e:
            self.__reply(command, reply=str(e), error=True, channel=channel)
            logging.exception('Failed to execute command %s', command)

    def __reply(self, msg, reply, channel, error=None):
        reply_msg = message().header('reply-for', msg.id).\
            header('type', 'reply').\
            value('error', error).value('reply', reply).build()
        ser_msg = serialize(reply_msg)
        channel.send(ser_msg)

    def _merge_apps(apps, napps, node):
        for napp in napps:
            app = apps.get(napp.name)
            if not app:
                app = {
                    'name': napp.name,
                    'description': napp.description,
                    'command': napp.command,
                    'params': napp.params,
                    'needs': napp.needs,
                    'nodes': []
                }
                apps[napp.name] = app
            app['nodes'].append(node)

    def get_available_apps(self):
        apps = {}
        Node._merge_apps(apps, self.get_apps(), self.get_node_info())

        for node_name, node_info in self.sync_manager.known_nodes.items():
            if node_name is not self.node_id:
                Node._merge_apps(apps, node_info.apps, node_info)
        return apps

    def get_stats(self):
        pass

    def get_apps(self):
        return [app for name, app in self.store.apps.items()]

    def query(self, q):
        pass

    def run_app(self, app_name):
        # find app
        # sort nodes by requirements
        apps = self.get_available_apps()

        app = apps.get(app_name)
        if not app:
            raise Exception('No such app %s' % app_name)

        ranked = Node._rank_nodes(app['needs'], app['nodes'])
        for ranked_node in ranked:
            try:
                return self._run_as_task(app, ranked[0])
            except Exception as e:
                logging.exception('Failed to run on node %s' % ranked_node)
        raise Exception('Failed to run app %s'%app_name)

    def _run_as_task(self, app, ranked_node):
        remote = ranked_node['node'] != self.node_id
        node = {'name': ranked_node['node']}

        if remote:
            node_info = self.sync_manager.get_node_info(ranked_node['node'])
            node['host'] = node_info.hostname
            node['port'] = node_info.data['ssh'].get('port') or 22
            node['ssh_user'] = node_info.data['ssh'].get('user') or 'root'

        task = task_for_app(app=app, remote=remote, node=node)
        return self.runner.run(task)

    def _rank_nodes(app_needs, nodes_info):
        m = max([v for k,v in app_needs.items()])
        W = {}
        for k, v in app_needs.items():
            W[k] = v/m

        return sorted([{'score': Node._calc_score(node_info.stats, W), 'stats': node_info.stats, 'node': node_info.name} for node_info in nodes_info], key=lambda x: x['score'], reverse=True)

    def _calc_score(stats, W):
        score = 0
        # CPU score
        score += Node._relevant_cpu_value(stats) * W['cpu']
        score += stats.memory['available'] * W['memory']
        score += stats.disk['ioload'] * (-W['disk'])
        # FIXME: add network score
        return score

    def _relevant_cpu_value(stats):
        # available bogomips = total bogomips - cpu usage * total bogomips = total bogomips * (1 - cpu usage)
        total_bogomips = stats.cpu['bogomips']['total']
        usage = stats.cpu['usage']
        return total_bogomips * (1 - usage)

    def start(self):
        if self.config.get('lock'):
            self.__lock()
        self.channel_manager = self._start_channel_manager_()
        self.log.info('Node %s started' % self.node_id)
        self._start_stats_tracker_()
        self._start_sync_manager_()
        self.__register_message_dispatcher__()
        self.__register_to_local()
        self.__register_command_handlers()

    def stop(self):
        self.log.debug('node stop')
        if self.stats_tracker:
            self.stats_tracker.stop_tracking()
            self.log.info('Statistics tracking has stopped')
        if self.aio_server:
            self.aio_server.stop()
            self.log.info('Async I/O Server notified to stop')
        if self.sync_manager:
            self.sync_manager.stop()
        if self.lock:
            self.lock.unlock()
        self.runner.shutdown()
        self.log.debug('Runner stopped')

    def get_node_info(self):
        return NodeInfo(name=self.node_id, stats=self.stats_tracker.get_stats(),
                        apps=self.get_apps(), endpoint=self.aio_server.get_server_endpoint(),
                        hostname=self.stats_tracker.hostname)


class NodeInfo:
    def __init__(self, name=None, stats=None, apps=None, endpoint=None, hostname=None, data=None):
        self.name = name
        self.stats = stats
        self.apps = apps
        self.endpoint = endpoint
        self.hostname = hostname
        self.data = data or {}


def node_info_from_dict(node_dict):
    apps = []
    dapps = node_dict.get('apps') or []

    for dapp in dapps:
        app = deserialize_dict(dapp, App)
        apps.append(app)

    dstats = node_dict.get('stats') or {}
    stats = deserialize_dict(dstats, SystemStats)
    node = deserialize_dict(node_dict, NodeInfo)
    node.apps = apps
    node.stats = stats
    return node


class RandomBuffer:

    def __init__(self, nodes_dict):
        self.nodes_dict = nodes_dict
        self.buffer = []
        self._shuffle_()

    def next(self, n):
        n = int(ceil(n))
        while len(self.buffer) < n:
            self._shuffle_()
        shuffled = self.buffer[0:n]
        self.buffer = self.buffer[n:]
        return shuffled

    def _shuffle_(self):
        nodes = [name for name, node in self.nodes_dict.items()]
        random.shuffle(nodes)
        self.buffer = self.buffer + nodes


class SyncManager:

    def __init__(self, node, channel_manager, event_processor, sync_interval=60000, sync_percent=0.3):
        self.node = node
        self.channel_manager = channel_manager
        self.event_processor = event_processor
        self.sync_percent = sync_percent
        self.known_nodes = {}
        self.random_buffer = RandomBuffer(self.known_nodes)
        self.sync_timer = IntervalTimer(offset=sync_interval, interval=sync_interval, target=self.sync_random_nodes)

    def _on_message_(self, msg_str, channel):
        msg = deserialize(msg_str, as_type=Message)
        #print('Got message [%s]' % msg.__dict__)
        if msg.headers.get('type') == 'sync-message':
            self._on_sync_message_(msg)

    def _on_sync_message_(self, msg):
        #print('Got sync message -> %s' % msg)
        #print(' : From node %s(%s)' % (msg.data['node']['name'], msg.data['node']['endpoint']))
        nodes = [node_info_from_dict(msg.data['node'])]

        known_nodes = [node_info_from_dict(node) for node in msg.data['known_nodes']]
        nodes = nodes + known_nodes
        self._merge_nodes_list_(nodes)

    def _merge_nodes_list_(self, nodes):
        for node in nodes:
            if node.name == self.node.node_id:
                continue
            if not self.known_nodes.get(node.name):
                self.known_nodes[node.name] = node
                logging.info('Node %s has joined' % node.name)
                self._print_known_nodes_()
            self._merge_node_(node)

    def _print_known_nodes_(self):
        logging.debug(' Members: ')
        for name, node in self.known_nodes.items():
            logging.debug('   %s[%s]' % (name, node.endpoint))

    def _merge_node_(self, node):
        existing = self.known_nodes[node.name]
        self.known_nodes[node.name] = node
        if existing.endpoint != node.endpoint:
            self.channel_manager.close_channel(existing.endpoint)

    def _on_closed_channel_(self, channel):
        to_remove = []
        for name, node in self.known_nodes.items():
            if node.endpoint == channel.to_url:
                to_remove.append(name)

        for name in to_remove:
            del self.known_nodes[name]
            logging.info('Node %s has probably left' % name)

    def start(self):
        self.sync_timer.start()

        self.channel_manager.on('channel.closed', self._on_closed_channel_)
        self.channel_manager.on('channel.data', self._on_message_)

    def stop(self):
        self.sync_timer.cancel()
        self.channel_manager.remove_listener('channel.closed', self._on_closed_channel_)

    def register_node(self, node):
        if self.known_nodes.get(node.name):
            self._merge_node_(node)
        else:
            self.known_nodes[node.name] = node

    def unregister_node(self, node):
        pass

    def sync_random_nodes(self):
        nodes = self.random_buffer.next(len(self.known_nodes) * self.sync_percent)

        for name in nodes:
            if self.known_nodes.get(name):
                node = self.known_nodes[name]
                logging.debug('Sync with %s [%s]' % (name, node.endpoint))
                self.channel_manager.send(to_url=node.endpoint, data=serialize(self.get_sync_message(), indent=2))

    def sync_one_node(self, node, this_node_info):
        pass

    def get_sync_message(self):
        return message().value('node', self.node.get_node_info()).\
            value('known_nodes', [n for k, n in self.known_nodes.items()]).\
            header('type', 'sync-message').build()

    def get_node_info(self, node):
        node_info = self.known_nodes.get(node)
        if not node_info:
            raise Exception('Unknown node %s' % node)
        return node_info


NODE_LOCK_FILE_PATH = '/tmp/troup.node.lock'


def local_node_lock_file(node_info=None):
    return this_process_info_file(NODE_LOCK_FILE_PATH, info=node_info, create=True)


def check_local_node_lock():
    return open_process_lock_file(NODE_LOCK_FILE_PATH)


def read_local_node_lock():
    return open_process_lock_file(NODE_LOCK_FILE_PATH, read_only=True)



import sys

sys.path.append('..')
print(sys.path)
import troup
from troup.distributed import Promise

import time
import threading

promise = Promise()

def producer():
    time.sleep(2)
    promise.complete(result="Yay!")
    print('Produced result')

def consumer():
    result = promise.result
    print('Got result: %s' % result)

tp = threading.Thread(target=producer)
tc = threading.Thread(target=consumer)

tc.start()
tp.start()

time.sleep(5)
print('exit')



# Copyright 2016 Pavle Jonoski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from argparse import ArgumentParser
from troup.node import Node
import logging


def configure_node_parser():
    parser = ArgumentParser(prog='troup', description='Run single node')
    
    # Node
    parser.add_argument('--node', help='Node ID')
    parser.add_argument('--neighbours', default='', nargs='+', help='Neighbour nodes')
    
    # Async IO server props
    parser.add_argument('--host', default='', help='Async IO server hostname')
    parser.add_argument('--port', default=7000, help='Async IO server port')
    
    # Store
    parser.add_argument('--storage-root', default='.data', help='Root path of the storage directory')
    
    # System statistics
    parser.add_argument('--stats-update-interval', default=30000, help='Statistics update interval in milliseconds')
    
    parser.add_argument('--log-level', '-l', default='info', help='Logging level')

    parser.add_argument('--lock', action='store_true', help='Write node info in global lock file')

    parser.add_argument('--debug', action='store_true', help='Activate the debug command-line interactive interface')

    parser.add_argument('-v', '--version', action='store_true', help='Print version and exit')
    
    return parser
    

def run_node():
    import signal
    
    parser = configure_node_parser()
    args = parser.parse_args()
    
    if args.version:
        from troup.metadata import __version__
        print(__version__)
        return
        
    logging.basicConfig(level=getattr(logging, args.log_level.upper()))
    
    config = {
        'store': {
            'path': args.storage_root
        },
        'server': {
            'hostname': args.host,
            'port': args.port
        },
        'stats': {
            'update_interval': args.stats_update_interval
        },
        'neighbours': args.neighbours,
        'lock': args.lock
    }
    node = Node(node_id=args.node, config=config)
    
    def handle_node_shutdown(signal, frame):
        node.stop()
    
    signal.signal(signal.SIGINT, handle_node_shutdown)

    if args.debug:
        from troup.debug import run_debug_cli
        run_debug_cli()

    node.start()
    
    return node


__author__ = 'pavle'


import sys

sys.path.append('..')

from troup.client import CommandAPI, client_to_local_node

cmd = CommandAPI(channel_client=client_to_local_node())

promise = cmd.send(CommandAPI.task('LocalProcess', {
    'directory': '/home/pavle/projects',
    'executable': '/usr/bin/xterm',
    'args': []
}, track_out=True))
print(promise)
print('cmd send')
try:
    result = promise.result
    print('Result -> %s' % str(result))
finally:
    cmd.shutdown()
    print('done')



class Test:
    def __init__(self):
        self.prop1 = None
        raise Exception('Interrupt init')
    def __del__(self):
        print('instance deleted')
        print(self.prop1)

class Test2(Test):
    pass

Test2()



class A:
    def __init__(self):
        print('A init')

class M1(A):
    def __init__(self):
        super(M1, self).__init__()
        print('M1 init')

class M2(A):
    def __init__(self):
        super(M2, self).__init__()
        print('M2 init')

class R1(M2, M1):
    def __init__(self):
        super(R1, self).__init__()
        print('R1 init')
        
r = R1()



# Copyright 2016 Pavle Jonoski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from troup.apps import App

__author__ = 'pavle'


class Store:

    def add_app(self, app):
        pass

    def remove_app(self, app_name):
        pass

    def update_app(self, app):
        pass

    def find_app(self, app_name):
        pass

    def search_apps(self, query):
        pass

    def store_settings(self, settings):
        pass

    def get_settings(self):
        pass

    def get_setting(self, setting_name):
        pass

    def set_setting(self, name, value):
        pass


class InMemorySyncedStore(Store):

    APPS_FILE = 'apps.json'
    SETTINGS_FILE = 'settings.json'

    def __init__(self, root_path, apps_file=None, settings_file=None):
        self.root_path = root_path
        self.apps = {}
        self.settings = {}
        self.apps_file = apps_file or InMemorySyncedStore.APPS_FILE
        self.settings_file = settings_file or InMemorySyncedStore.SETTINGS_FILE
        self.__check_storage_files__()
        self.load_from_file()

    def __check_storage_files__(self):
        if not os.path.exists(self.root_path):
            os.makedirs(self.root_path, 0o755)
        self.__check_storage_file__(self.apps_file)
        self.__check_storage_file__(self.settings_file)

    def __check_storage_file__(self, f_path):
        if not self.__exists_and_is_file__(f_path):
            with open(self.__to_path__(f_path), 'w') as file:
                file.write('{}')

    def __exists_and_is_file__(self, path):
        path = self.__to_path__(path)
        return os.path.exists(path) and os.path.isfile(path)

    def __load_apps__(self):
        apps = {}
        apps_str = self.__load__(self.__to_path__(self.apps_file))
        apps_json = json.loads(apps_str)
        for app_id, app_json in apps_json.items():
            apps[app_id] = self.__get_app__(app_json)
        return apps

    def __get_app__(self, app_json):
        return App(name=app_json['name'], description=app_json.get('description'),\
            command=app_json['command'], params=app_json.get('params'),\
            needs=app_json.get('needs'))

    def __load__settings__(self):
        return json.loads(self.__load__(self.__to_path__(self.apps_file)))

    def __to_path__(self, file_name):
        return os.path.join(self.root_path, file_name)

    def __load__(self, path):
        if os.path.exists(path) and os.path.isfile(path):
            with open(path) as file:
                return file.read()
        raise Exception('Unable to read: %s. File does not exist or is not a file' % path)

    def __store__(self, path, data):
        path = self.__to_path__(path)
        with open(path, 'w') as file:
            file.write(data)

    def load_from_file(self):
        self.apps = self.__load_apps__()
        self.settings = self.__load__settings__()

    def ___store_apps___(self):
        apps_json = json.dumps(self.apps, cls=DictEncoder)
        self.__store__(self.apps_file, apps_json)

    def __store_settings__(self):
        settings_json = json.dumps(self.settings)
        self.__store__(self.settings_file, settings_json)

    def sync(self):
        self.___store_apps___()
        self.__store_settings__()

    def add_app(self, app):
        self.apps[app.name] = app
        self.___store_apps___()

    def remove_app(self, app_name):
        if self.apps.get(app_name):
            del self.apps[app_name]
            self.___store_apps___()

    def update_app(self, app):
        self.apps[app.name] = app
        self.___store_apps___()

    def find_app(self, app_name):
        return self.apps.get(app_name)

    def search_apps(self, query):
        apps = []
        ql = query.lower()
        for name, app in self.apps.items():
            if ql in name.lower() or (app.description and ql in app.description.lower()):
                apps.append(app)
        return apps

    def store_settings(self, settings):
        self.settings.update(settings)
        self.__store_settings__()

    def get_settings(self):
        return self.settings

    def get_setting(self, setting_name):
        return self.settings.get(setting_name)

    def set_setting(self, name, value):
        self.settings[name] = value
        self.__store_settings__()


class DictEncoder(json.JSONEncoder):
    
    def default(self, o):
        return o.__dict__



# Copyright 2016 Pavle Jonoski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = 'pavle'

import threading


def print_threads(cmd):
    for th in threading.enumerate():
        print(th)


def print_help(cmd):
    for cmd_name, entry in DBG_HANDLERS.items():
        desc, hnd = entry
        print('%10s - %s' % (cmd_name, desc))


def quit_cli(cmd):
    raise ExitCliException()

class ExitCliException(Exception):
    pass

DBG_HANDLERS = {
    'pt': ('Print Threads', print_threads),
    '?': ('Prints this help message', print_help),
    'help': ('Prints this help message', print_help),
    'q': ('Quits the debug cli', quit_cli),
    'quit': ('Quits the debug cli', quit_cli)
}


def debug_cli():

    while True:
        cmd = input('(debug cli): ')
        cmd = cmd.lower().strip()

        if not cmd:
            continue

        description, handler = DBG_HANDLERS.get(cmd)
        if not handler:
            print('Unrecognized command %d. Type ? for help.')
            continue
        try:
            handler(cmd)
        except ExitCliException:
            break
        except Exception as e:
            print('An error occurred: %s' % e)


def run_debug_cli():
    threading.Thread(target=debug_cli).start()



import threading
import time

class Consumer(threading.Thread):
    
    def __init__(self, event):
        super(Consumer, self).__init__()
        self.event = event
    
    def run(self):
        print('%s: waiting for event...'%self.name)
        self.event.wait()
        print('%s: got event'%self.name)

class Provider(threading.Thread):
    def __init__(self, event):
        super(Provider, self).__init__()
        self.event = event
        self.timeout = 1000
    
    def run(self):
        print('%s: sleeping for %d ms'%(self.name, self.timeout) )
        time.sleep(self.timeout/1000)
        print('%s: triggering now!' % self.name)
        self.event.set()

ev = threading.Event()

c1 = Consumer(ev)
c2 = Consumer(ev)
c3 = Consumer(ev)

p1 = Provider(ev)

c1.start()
c2.start()
c3.start()
p1.start()

c1.join()
c2.join()
c3.join()
p1.join()





# Copyright 2016 Pavle Jonoski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from troup.infrastructure import OutgoingChannelOverWS
from troup.distributed import Promise
from troup.threading import IntervalTimer
from troup.node import read_local_node_lock
from troup.messaging import message, serialize, deserialize, Message

from threading import Thread
from datetime import datetime, timedelta


class CallbackWrapper:
    def __init__(self, callback, valid_for, promise=None, created_on=None):
        self.callback = callback
        self.valid_for = valid_for
        self.created_on = created_on or datetime.now()
        self.promise = promise or Promise()

    def check_expired(self):
        if datetime.now() > (timedelta(milliseconds=self.valid_for) + self.created_on):
            self.promise.complete(error='Timeout', result=Exception('Timeout'))

    def execute_callback(self, result):
        if self.callback:
            try:
                self.callback(result)
            except Exception as e:
                print('Woops')
        self.promise.complete(result=result)


class ChannelClient:
    def __init__(self, nodes_specs=None, reply_timeout=5000, check_interval=5000):
        self.nodes_ref = {}
        self.channels = {}
        self.callbacks = {}
        self.reply_timeout = reply_timeout
        self.check_interval = check_interval
        self.maintenance_timer = self.__build_timer()
        self.__build_nodes_refs__(nodes_specs)

    def __build_nodes_refs__(self, nodes_specs):
        for spec in nodes_specs:
            parsed = spec.partition(':')
            self.nodes_ref[parsed[0]] = parsed[2]

    def __build_timer(self):
        timer = IntervalTimer(interval=self.check_interval, offset=self.check_interval,
                              target=self.__check_expired_callbacks)
        timer.start()
        return timer

    def __check_expired_callbacks(self):
        for msgid, wrapper in self.callbacks.items():
            wrapper.check_expired()

    def __reg_wrapper(self, message, callback):
        wrapper = CallbackWrapper(callback=callback, valid_for=5000)
        self.callbacks[message.id] = wrapper
        return wrapper

    def __on_channel_data(self, data, channel):
        msg = deserialize(data, Message)
        if msg.headers.get('type') == 'reply':
            self.__process_reply(msg)

    def __process_reply(self, reply):
        id = reply.headers.get('reply-for')
        if not id:
            raise Exception('Invalid reply %s' % reply)
        wrapper = self.callbacks.get(id)
        if wrapper:
            if reply.data.get('error'):
                wrapper.promise.complete(error=reply.data.get('reply'))
            else:
                wrapper.promise.complete(result=reply.data.get('reply'))

    def send_message(self, message, to_node=None, on_reply=None):
        def reply_callback_wrapper(*args, **kwargs):
            if on_reply:
                on_reply(*args, **kwargs)
        wrapper_promise = Promise()
        def do_send():
            promises = []
            if to_node:
                promise = self.send_message_to_node(message, to_node, reply_callback_wrapper)
                promises.append(promise)
            else:
                for name, node in self.nodes_ref.items():
                    promise = self.send_message_to_node(message, name, reply_callback_wrapper)
                    promises.append(promise)
            results = []
            for p in promises:
                results.append(p.result)
            if len(results) == 1:
                wrapper_promise.complete(result=results[0])
            else:
                wrapper_promise.complete(result=True)

        def run_in_thread():
            try:
                do_send()
            except Exception as e:
                wrapper_promise.complete(error=e)

        Thread(target=run_in_thread).start()
        return wrapper_promise

    def send_message_to_node(self, message, node, on_reply):
        channel = self.get_channel(node)
        wrapper = self.__reg_wrapper(message=message, callback=on_reply)
        ser_message = serialize(message)
        channel.send(ser_message)
        return wrapper.promise

    def get_channel(self, for_node):
        channel = self.channels.get(for_node)
        if not channel:
            channel = self.build_channel(for_node)
        return channel

    def build_channel(self, for_node):
        ref = self.nodes_ref.get(for_node)
        if not ref:
            raise Exception('Unknown node reference [%s]' % for_node)
        return self.create_channel(for_node, ref)

    def create_channel(self, node_name, reference):
        chn = OutgoingChannelOverWS(node_name, reference)

        def on_data(data):
            #print('DATA %s' % data)
            self.__on_channel_data(data, channel=chn)

        chn.register_listener(on_data)
        chn.open()
        self.channels[node_name] = chn
        return chn

    def shutdown(self):
        for name, channel in self.channels.items():
            channel.close()
        self.maintenance_timer.cancel()


def client_to_local_node():
    lock = read_local_node_lock()
    client = ChannelClient(nodes_specs=['%s:%s' % (lock.get_info('name'), lock.get_info('url'))])
    return client


class CommandAPI:
    def __init__(self, channel_client):
        self.channel_client = channel_client

    def send(self, command, to_node=None, on_reply=None):
        return self.channel_client.send_message(message=command, to_node=to_node, on_reply=on_reply)

    def monitor(self, command_ref):
        pass

    def command(name, data):
        return message(data=data).header('type', 'command').header('command', name).build()

    def task(type, data, ttl=None, track_out=False, buffer=None):
        return message().header('type', 'task').header('ttl', ttl).\
            header('task-type', 'process').header('process-type', type).\
            header('consume-out', track_out).header('buffer-size', buffer).\
            value('process', data).build()

    def shutdown(self):
        self.channel_client.shutdown()


if __name__ == '__main__':
    from argparse import ArgumentParser
    from json import loads, dumps

    parser = ArgumentParser(prog="troup.client", description="Low level troup system client")

    parser.add_argument('--node', help='Node connection URL.', default='ws://localhost:7000')
    parser.add_argument('-t', '--type', default='command', help='Message type. May be "command" or "task".')
    parser.add_argument('-d', '--data', help='Message data. This is usually a JSON string.')
    parser.add_argument('-H', '--header', nargs='+', help='Message headers in the form HEADER_NAME=VALUE.')

    parser.add_argument('--reply-timeout', default=5000, help='Message reply timeout in milliseconds.')
    parser.add_argument('--check-interval', default=1000, help='Interval to check for timeouts in milliseconds.')

    parser.add_argument('-c', '--command', help='The command name. Used only when type is "command".')

    parser.add_argument('-v', '--verbose', help='Be more verbose.', action='store_true')

    parser.add_argument('--as-json', action='store_true',
                        help='Try to serialize the result as JSON and print it on stdout.')

    args = parser.parse_args()

    def printout(*arguments):
        if args.verbose:
            print(*arguments)

    data = args.data or '{}'
    data = loads(data)

    mb = message(data=data)
    for header in args.header or []:
        try:
            header_name, value = header.split('=')
            mb.header(header_name, value)
        except Exception as e:
            raise Exception('Invalid header value %s' % header) from e

    mb.header('type', args.type)

    if args.type == 'command' and args.command:
        mb.header('command', args.command)

    try:
        message = mb.build()

        channel_client = ChannelClient(nodes_specs=['TARGET:%s'%args.node], reply_timeout=args.reply_timeout,
                                       check_interval=args.check_interval)

        api = CommandAPI(channel_client=channel_client)

        printout('Sending to ', args.node)

        promise = api.send(command=message, to_node='TARGET')

        result = promise.result

        if args.as_json:
            print(dumps(result))
        else:
            print(result)
    finally:
        if api:
            api.shutdown()


import sys
import time
import threading

sys.path.append('..')

print(sys.path)

from troup.infrastructure import OutgoingChannelOverWS

oc = OutgoingChannelOverWS('tc', 'ws://localhost:8910')


class RunAlone(threading.Thread):
    def run(self):
        oc.connect()
        

a = RunAlone()
a.start()
time.sleep(10)
print('Done')






# Copyright 2016 Pavle Jonoski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

class Observable:
    
    def __init__(self):
        self.listeners = {}
    
    def on(self, event, handler):
        listeners = self.listeners.get(event)
        if not listeners:
            listeners = self.listeners[event] = []
        listeners.append(handler)
    
    def trigger(self, event, *args):
        listeners = self.listeners.get(event)
        if listeners:
            for listener in listeners:
                listener(*args)
    
    def remove_listener(self, event, listener):
        listeners = self.listeners.get(event)
        if listeners:
            listeners.remove(listener)



import sys

sys.path.append('..')

from troup.client import CommandAPI, client_to_local_node, ChannelClient

#cc = ChannelClient(nodes_specs=['RPI:ws://192.168.2.128:7000'])
cc = client_to_local_node()

cmd = CommandAPI(channel_client=cc)

promise = cmd.send(CommandAPI.command('info', {}))
print(promise)
print('cmd send')
try:
    result = promise.result
    print('Result -> %s' % str(result))
finally:
    cmd.shutdown()
    print('done')



# Copyright 2016 Pavle Jonoski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from threading import Timer, Condition


class AtomicOperation:

    def __init__(self):
        self.lock = Condition()

    def op(self):
        with self.lock:
            return self.atomic_op()

    def atomic_op(self):
        pass


class AtomicIncrement(AtomicOperation):

    def __init__(self, initial_value=0):
        self._int_val = initial_value
        super(AtomicIncrement, self).__init__()

    def atomic_op(self):
        self._int_val += 1

    def inc(self):
        return self.op()

    @property
    def value(self):
        return self._int_val

    def __repr__(self):
        return '%d' % self._int_val

    def __str__(self):
        return self.__repr__()

_id_inc = AtomicIncrement()


def _next_id(pref):
    _id_inc.inc()
    return '%s-%s' %(pref, _id_inc)


class IntervalTimer:
    
    def __init__(self, interval, offset=0, target=None, name=None):
        self.target = target
        self.timer = None
        self.running = False
        self.offset = offset
        self.interval = interval
        self.first = True
        self.name = name or _next_id('IntervalTimer')
        
    def start(self):
        if self.running:
            return
        interval = self.interval
        if self.first:
            interval = interval + self.offset
            self.first = False
        self._run_()

    def cancel(self):
        if self.timer:
            self.timer.cancel()
            self.running = False

    def _run_(self):
        self.running = True
        self.timer = Timer(self.interval/1000, self._run_)
        self.timer.name = self.name
        self.timer.start()
        self.run()
        self.running = False
    
    def run(self):
        if self.target:
            self.target()


"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""
from setuptools import setup, find_packages
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

from troup.metadata import __version__

setup(
    name='troup',
    version=__version__,
    description='Troup system',
    long_description=long_description,
    url='https://github.com/troup-system/troup',

    author='Pavle Jonoski',
    author_email='jonoski.pavle@gmail.com',
    license='http://www.apache.org/licenses/LICENSE-2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 2 - Pre-Alpha',

        'Intended Audience :: Developers',
        'Topic :: System :: Distributed Computing',
        'Topic :: Utilities',

        'License :: OSI Approved :: Apache Software License',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    keywords='development distributed computing',

    packages=find_packages(exclude=['contrib', 'docs', 'tests']),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    install_requires=['ws4py', 'psutil', 'nose'],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
    },

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
    entry_points={
        'console_scripts': [
            'sample=troup.setup:main',
        ],
    },
)



# Copyright 2016 Pavle Jonoski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
from functools import wraps

class expect_content:
    def __init__(self, file=None, expect_file=None, content_type='text'):
        self.file = file
        self.expect_file = expect_file
        self.content_type = content_type
        self.method = None

    def wrapper(self, *args, **kwargs):
        if not self.method:
            return
        self.method(*args, **kwargs)
        with open(self.file) as given:
            with open(self.expect_file) as expected:
                self.compare(given, expected)

    def compare(self, given, expected):
        if self.content_type == 'text':
            compare_text_content(given, expected)
        elif self.content_type == 'json':
            compare_json_content(given, expected)
        else:
            raise Exception('Invalid content type %s' % self.content_type)

    def __call__(self, method):
        self.method = method

        @wraps(method)
        def wrapper(*args, **kwargs):
            return self.wrapper(*args, **kwargs)

        return wrapper

class load_content:

    def __init__(self, path, data_type='json', bound_method=True):
        self.path = path
        self.type = data_type
        self.bound_method = True

    def __call__(self, method):
        @wraps(method)
        def wrapper(*args, **kwargs):
            content = self._load_path()
            nargs = (*args, content)
            method(*nargs, **kwargs)

        return wrapper

    def _load_path(self):
        load_method_name = 'load_%s' % self.type.lower()
        if hasattr(self, load_method_name):
            load_method = getattr(self, load_method_name)
            return load_method()
        else:
            return self.load_default()

    def load_json(self):
        return load_json(self.path)

    def load_default(self):
        with open(self.path) as fs:
            return fs.read()


def compare_text_content(gf, ef):
    given_text = gf.read()
    expected_text = ef.read()
    if given_text != expected_text:
        raise Exception('Text compare failed. Expected [%s] but was given [%s]' %(expected_text, given_text))


def compare_json_content(gf, ef):
    given_json = json.loads(gf.read())
    expected_json = json.loads(ef.read())
    dict_compare(given_json, expected_json)


def list_compare(given, expected):
    if not isinstance(given, list):
        raise Exception('The given object is not list. Expected to get list.')
    if len(given) != len(expected):
        raise Exception('Compare failed. Expected to get list with size %d, but was given list with size %d.' %(len(expected), len(given)))

    i = 0
    for item in expected:
        expected_item = expected[i]
        try:
            if isinstance(item, dict):
                dict_compare(item, expected_item)
            elif isinstance(item, list):
                list_compare(item, expected_item)
            else:
                if item != expected_item:
                    raise Exception('Expected [%s] but was given [%s]' %(expected_item, item))
        except Exception as e:
            raise Exception('Compare failed. List element at index %d: %s' % (i, str(e))) from e
        i += 1

def dict_compare(given, expected):
    if not isinstance(given, dict):
        raise Exception('The given object is not dict. Expected to get dictionary.')
    for key, expected_value in expected.items():
        value = given.get(key)
        try:
            if isinstance(value, dict):
                dict_compare(value, expected_value)
            elif isinstance(value, list):
                list_compare(value, expected_value)
            else:
                if value != expected_value:
                    raise Exception('Expected [%s] but was given [%s]' %(expected_value, value))
        except Exception as e:
            raise Exception('Compare failed. Element for key [%s]: %s' % (key, str(e))) from e


def load_json(file_path):
    with open(file_path) as fl:
        return json.loads(fl.read())



# Copyright 2016 Pavle Jonoski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import uuid4
from concurrent.futures import ThreadPoolExecutor
from threading import Thread
from collections import deque
import logging
from functools import reduce

from troup.process import LocalProcess, SSHRemoteProcess
from troup.threading import IntervalTimer
from datetime import datetime, timedelta

class TaskException(Exception):
    pass


class TaskRunException(TaskException):
    pass


class TaskRun:
    
    CREATED = 'CREATED'
    RUNNING = 'RUNNING'
    STOPPING = 'STOPPING'
    DONE = 'DONE'
    ERROR = 'ERROR'
    
    def __init__(self, task, future=None, run_id=None):
        self.task = task
        self.status = TaskRun.CREATED
        self.result = None
        self.error = None
        self.future = future
        self.start_time = None
        self.ttl = task.ttl
        self.id = run_id or task.id or str(uuid4())

    def start(self):
        if self.status is not TaskRun.CREATED:
            raise TaskRunException('Failed to start task: invalid Status %s' % self.status)
        try:
            self.status = TaskRun.RUNNING
            self.start_time = datetime.now()
            self.do_start()
            if self.status is TaskRun.RUNNING:
                self.stop()
        except Exception as e:
            logging.exception('Failed to execute task')
            self.status = TaskRun.ERROR
            self.error = e

    def stop(self, reason=None):
        if self.status is not TaskRun.RUNNING:
            raise TaskRunException('Failed to stop task: invalid Status %s' % self.status)
        self.status = TaskRun.STOPPING
        try:
            self.do_stop(reason)
            self.status = TaskRun.DONE
        except Exception as e:
            logging.exception('Failed to stop task')
            self.error = e
            self.status = TaskRun.ERROR

    def do_start(self):
        self.task.run()

    def do_stop(self, reason=None):
        self.task.stop(reason)


class TasksRunner:

    def __init__(self, max_workers=3):
        self.tasks = {}
        self.executor = ThreadPoolExecutor(max_workers=max_workers)
        self.checker = IntervalTimer(1000, offset=1000, target=self._check_tasks, name='TasksRunnerMaintenanceTimer')
        self.checker.start()

    def _check_tasks(self):
        to_remove = []
        for id, task_run in self.tasks.items():
            if task_run.status is TaskRun.DONE or task_run.status is TaskRun.ERROR:
                to_remove.append(task_run)
        for task_run in to_remove:
            self.__remove_task(task_run)

    def run(self, task):
        if self.tasks.get(task.id):
            raise TaskRunException('Task already running %s' % str(task))

        task_run = TaskRun(task=task)
        self.tasks[task.id] = task_run

        def start_task():
            print('Running task [%s]' % task_run.id)
            task_run.start()
            print('Run and done - task [%s]' % task_run.id)

        def on_done(*args):
            if self.tasks.get(task.id):
                run = self.tasks[task.id]
                run.result = run.future.result()
                if not run.ttl:
                    del self.tasks[task.id]

        try:
            future = self.executor.submit(start_task)
            task_run.future = future
            future.add_done_callback(on_done)
            return task_run
        except Exception as e:
            del self.tasks[task.id]
            raise TaskRunException('Failed to schedule task run for %s' % str(task)) from e

    def stop(self, task_id, wait=False, timeout=None):
        if not self.tasks.get(task_id):
            raise TaskException('No task with id %s' % task_id)
        task = self.tasks[task_id]
        try:
            task.stop()
            if wait:
                task.future.result(timeout=timeout)
        except Exception as e:
            raise TaskException('Failed to stop task %s' % str(task)) from e
        finally:
            if self.tasks.get(task_id):
                del self.tasks[task_id]

    def shutdown(self):
        self.checker.cancel()
        self.executor.shutdown(wait=True)

    def __get_task(self, task_id):
        task = self.tasks.get(task_id)
        if not task:
            raise TaskException('No task with id %s' % task_id)
        return task

    def __remove_task(self, task, force=False):
        if force or not task.ttl:
            del self.tasks[task.id]
        elif task.ttl > 0:
            td = datetime.now() - task.start_time
            ttl = timedelta(microseconds=task.ttl*1000)
            if td > ttl:
                del self.tasks[task.id]

    def clear(self, task_id):
        task_run = self.__get_task(task_id)
        if task_run is TaskRun.RUNNING:
            raise TaskException('Task is running')
        self.__remove_task(task=task_run, force=True)

    @property
    def stats(self):
        result = {
            'total': len(self.tasks)
        }
        running = 0
        for id, run in self.tasks.items():
            result[id] = {
                'id': id,
                'status': run.status,
                'since': run.start_time
            }
            if run.status is TaskRun.RUNNING:
                running += 1
        result['running'] = running
        return result


class Task:
    def __init__(self, task_id=None, ttl=None):
        self.id = task_id or str(uuid4())
        self.ttl = ttl
        self.result = None
    
    def run(self, context=None):
        pass

    def stop(self, reason=None):
        pass

    def __repr__(self):
        return '<%s.%s with id %s>' %(self.__class__.__module__, self.__class__.__name__, self.id)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if other is None:
            return False
        return type(self) == type(other) and self.id == other.id

    def __hash__(self):
        return self.id.__hash__()


class CodeTask(Task):
    def __init__(self, task_id, code, exec_type, ttl=None, data=None):
        super(CodeTask, self).__init__(task_id, ttl)
        self.code = code
        self.type = exec_type
        self.data = data
        
    def run(self, context):
        if not context.get('locals'):
            context['locals'] = {}
        if not context.get('globals'):
            context['globals'] = globals()
    
        return self.__exec__(context['locals'], context['globals'])
    
    def __exec__(self, task_globals, task_locals):
        if not task_locals.get('result'):
            task_locals['result'] = self.__set_result__
        result = self.exec(task_globals, task_locals)
        if result is not None:
            self.result = result
        return self.result
    
    def exec(self, task_globals, task_locals):
        pass
    
    def __set_result__(self, result):
        self.result = result


class FunctionBytecodeTask(CodeTask):
    
    def __init__(self, task_id, bytecode, fn_args, fn_kwargs, data, ttl=None):
        super(FunctionBytecodeTask, self).__init__(task_id, code=bytecode, data=data, exec_type="FunctionBytecode", ttl=ttl)
        self.fn_args = fn_args
        self.function = None
        
    def __build_function(self, fn_locals, fn_globals):
        #FunctionType()
        pass

    def exec(self, fn_locals, fn_globals):
        #self.function = self.__build_function()
        pass


class ProcessTaskException(TaskException):
    pass


class LocalProcessTask(Task):

    def __LocalProcessBuilder(id, data):

        executable = data['executable']
        args = data['args']
        cwd = data.get('directory')

        process = LocalProcess(id=id, name=executable, args=args, cwd=cwd)

        return process

    def __SSHProcessBuilder(id, data):
        executable = data['executable']
        host = data['host']
        port = data.get('port', '22')
        cwd = data.get('directory')
        forward_video = data.get('forward_video', False)
        forward_audio = data.get('forward_audio', False)
        compress_stream = data.get('compress_stream', False)
        ssh_user = data.get('ssh_user', '')
        args = data.get('args', [])

        process = SSHRemoteProcess(id=id, name=executable, cwd=cwd, target_host=host, target_port=port,
                                   forward_video=forward_video, forward_audio=forward_audio,
                                   compress_stream=compress_stream, ssh_user=ssh_user, args=args)

        return process

    PROCESS_BUILDERS = {
        'LocalProcess': __LocalProcessBuilder,
        'SSHProcess': __SSHProcessBuilder
    }

    def __init__(self, process_type, process_data, task_id=None, ttl=None,
                 consume_process_out=False, buffer_size=100000):
        super(LocalProcessTask, self).__init__(task_id=task_id, ttl=ttl)
        self.process = None
        self.__build_process(process_type, process_data)
        self.buffer_size = buffer_size
        self.consume_process_out = consume_process_out

        if self.consume_process_out:
            self._setup_process_consumers()

    def _setup_process_consumers(self):
        self._out_buffer = deque(maxlen=self.buffer_size)
        self._err_buffer = deque(maxlen=self.buffer_size)
        self._out_consumer = Thread(target=self._consume_out, name='Consumer:OUT:%s' % self.id)
        self._err_consumer = Thread(target=self._consume_err, name='Consumer:ERR:%s' % self.id)

    def _consume_out(self):
        try:
            while self.process and self.process.output and not self.process.output.closed:
                line = self.process.output.readline()
                print('[OUT]%s' % line)
                if not line:
                    break
                self._out_buffer.appendleft(line)
        except:
            logging.exception('Failed to read process output')

    def _consume_err(self):
        try:
            while self.process and self.process.error and not self.process.error.closed:
                line = self.process.error.readline()
                print('[ERR]%s' % line)
                if not line:
                    break
                self._out_buffer.appendleft(line)
        except:
            logging.exception('Failed to read process error')

    def _run_consumers(self):
        if self.consume_process_out:
            self._out_consumer.start()
            self._err_consumer.start()

    def __build_process(self, process_type, process_data):
        builder = LocalProcessTask.PROCESS_BUILDERS.get(process_type)
        if not builder:
            raise ProcessTaskException('Unsupported process type %s' % process_type)
        try:
            self.process = builder(self.id, process_data)
        except Exception as e:
            raise ProcessTaskException('Failed to build process of type %s' % process_type) from e

    def run(self, context=None):
        print('Executing process %s' % self.process)
        self.process.execute()
        if self.consume_process_out:
            self._run_consumers()
        print('Process started. Waiting...')
        returncode = self.process.wait()
        print('Process ended with code %d' % returncode)
        try:
            if returncode:
                msg = self._handle_error(returncode)
            else:
                self.result = self._collect_result()
                print('[[RESULT]]: ', self.result)
        finally:
            self.process.close_streams()

    def _collect_result(self):
        result = ''
        print('Track process out -> ', self.consume_process_out)
        if self.consume_process_out:
            result += reduce(lambda a, b: a + str(b, 'utf-8'), self._out_buffer, '')
        return result

    def _handle_error(self, returncode):
        message = 'code: %d' % returncode
        if self.consume_process_out:
            err_msg = reduce(lambda a, b: a + str(b, 'utf-8'), self._err_buffer, '')
            message += err_msg

        raise ProcessTaskException(message)

    def stop(self, reason=None):
        self.process.kill()


def __local_process_task_from_message(msg):
    process_type = msg.headers.get('process-type')
    if not process_type:
        raise ProcessTaskException('No process type specified')
    process_data = msg.data['process']
    task_id = msg.headers.get('task-id') or str(uuid4())
    ttl = int(msg.headers.get('ttl') or 0)
    buffer_size = msg.headers.get('buffer-size')
    consume_out = msg.headers.get('consume-out') or False
    return LocalProcessTask(process_type=process_type, process_data=process_data, task_id=task_id,
                            ttl=ttl, buffer_size=buffer_size, consume_process_out=consume_out)

__TASK_BUILDERS = {
    'process': __local_process_task_from_message
}


def build_task(msg):
    task_type = msg.headers.get('task-type')
    if not task_type:
        raise TaskException('Task type missing')
    builder = __TASK_BUILDERS.get(task_type)
    if not builder:
        raise TaskException('Cannot build task of type %s' % task_type)
    return builder(msg)


def task_for_app(app, remote=False, node=None, ttl=0, consume_output=False, buffer_size=1024*1024):
    process_type = 'SSHProcess' if remote else 'LocalProcess'
    process_data = {
        'executable': app['command']
    }

    if remote:
        process_data.update({
            'host': node['host'],
            'port': node['port'],
            'ssh_user': node['ssh_user'],
            'args': []
        })

    process_data.update(app['params'])

    if not process_data.get('args'):
        process_data['args'] = []

    task = LocalProcessTask(process_type=process_type, process_data=process_data, task_id=str(uuid4()), ttl=ttl,
                            consume_process_out=consume_output, buffer_size=buffer_size)
    return task




# Copyright 2016 Pavle Jonoski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = 'pavle'

import psutil
import os
import platform
import re

from troup.threading import IntervalTimer

class SystemStats:
    """Statistics values for the underlying system.
    
    Holds the values for a single measurement of the system performance.
    
    Attributes:
        cpu (dict): Holds the CPU values in a dictionary. The following values 
            are retrieved: 
            * "usage" (number): CPU usage. It is a number 0.0 - 1.0.
            * "per_cpu" (Array[number]): CPU usage per CPU. Reports the usage of each
                CPU separately in an Array.
            * "processors" (int): Number of processors present on the system.
            * "bogomips" (number): Bogus MIPS as reported by the Linux kernel (if available)
        memory (dict): Holds the values for the system memory:
            * "total" (int): total memory present in the system in Bytes.
            * "user" (int): Used memory (in Bytes).
            * "free" (int): Free memory (in Bytes).
        system (dict): System load and general system info:
            * "load" (Array[number]): System load as reported by the system. Usually this 
                is an array of three values - the system load for the last 5, 10 and 15 minutes.
            * "name" (str): Node name or the name of the machine.
            * "platform" (str): The underlying OS platform. May be "linux", "bsd" etc.
        disk (dict): Information about the disk and I/O of the system:
            * "ioload" (number): Normalized Input/Output load of the system - a number
                between 0.0 and 1.0. This is the average value for the disk usage in a 
                previous time interval.
    """

    def __init__(self):
        self.cpu = {'usage': 0.0, "per_cpu": [], 'processors': 0, 'bogomips': {'total': 0}}
        self.memory = {'total': 0, 'used': 0, 'available': 0}
        self.system = {'load': [0.0, 0.0, 0.0], 'name': '', 'platform': ''}
        self.disk = {'ioload': 0.0}
                

class StatsTracker:

    def __init__(self, period=1000):
        self.period = period
        self.cpu_usage = []
        self.cpu_usage_avg = psutil.cpu_percent()/100
        self.cpu_count = psutil.cpu_count()
        self.hostname = platform.node()
        self.platform = platform.system()
        self.periodic_update = IntervalTimer(interval=self.period, target=self.refresh_values)
        self.periodic_update.start()
        self.bogomips = get_bogomips()
    
    def refresh_values(self):
        cpu_usage = psutil.cpu_percent(percpu=True)
        self.cpu_usage = [usage/100 for usage in cpu_usage]
    
    def get_stats(self):
        stats = SystemStats()
        
        stats.cpu = self._get_cpu_stats_()
        stats.memory = self._get_mem_stats_()
        stats.system = self._get_system_stats_()
        stats.disk = self._get_disk_stats_()
        
        return stats
    
    def _get_disk_stats_(self):
        stats = {
            'ioload': 0.0
        }
        
        return stats
    
    def _get_cpu_stats_(self):
        stats = {
            'usage': self.cpu_usage_avg,
            'per_cpu': self.cpu_usage,
            'processors': self.cpu_count,
            'bogomips': self.bogomips
        }
        return stats
    
    def _get_mem_stats_(self):
        mem = psutil.virtual_memory()
        stats = {
            'total': mem.total,
            'used': mem.used,
            'available': mem.available
        }
        return stats
    
    def _get_system_stats_(self):
        stats = {}
        one, five, fifteen = os.getloadavg()
        stats['load'] = [one, five, fifteen]
        
        return stats
    
    def stop_tracking(self):
        self.periodic_update.cancel()



class CpuinfoParser:
    
    ENTRY_REGEX = '(?P<label>[\w\d-]+)\s?:\s?(?P<value>.+)'
    
    def __init__(self):
        self.processors = {}
        
        self._processor = None
        self._features = {}
        
    def parse(self, cpu_info_file):
        line = cpu_info_file.readline()
        
        while line:
            line = line.strip()
            if line:
                self._parse_line(line)
            line = cpu_info_file.readline()
        return self.processors
    
    def _parse_line(self, line):
        m = re.search(CpuinfoParser.ENTRY_REGEX, line)
        if m:
            label = m.group('label')
            value = m.group('value')
            if label == 'processor':
                if not self._processor:
                    self.processors[value] = self._features
                else:
                    self.processors[value] = self._features = {}
            self._features[label] = value
                
                

def get_bogomips():
    bogomips = {
        'total': 0,
        'cpus': {}
    }
    with open('/proc/cpuinfo') as cpuinfo:
        parser = CpuinfoParser()
        processors = parser.parse(cpuinfo)
        for processor, features in processors.items():
            bmps = features.get('bogomips')
            if bmps:
                bmps = float(bmps)
                bogomips['total'] += bmps
                bogomips['cpus'][processor] = bmps
    
    return bogomips

if __name__ == '__main__':
    import json
    import time
    
    tracker = StatsTracker(period=1000)
    
    while(True):
        s = tracker.get_stats()
        print(json.dumps(s.__dict__, indent=2))
        time.sleep(1)



class decorate:
    def __init__(self, arg1=None, arg2=None):
        self.arg1 = arg1
        self.arg2 = arg2
    
    def __call__(self, method):
        def wrapper():
            print('ARG1 = %s, ARG2 = %s' % ( str(self.arg1), str(self.arg2)) )
            method()
        return wrapper
@decorate('the decoration', 34)
def test():
    print('test method')

test()



# Copyright 2016 Pavle Jonoski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = 'pavle'

# https://docs.python.org/3.5/library/subprocess.html


from subprocess import Popen, PIPE
from os import path, getpid, remove
import json
import logging


class Process:

    def __init__(self, id, name, args=None):
        self.name = name
        self.id = id
        self.args = args

    def execute(self):
        pass

    def send_signal(self, signal):
        pass

    def get_returncode(self):
        pass


class LocalProcess(Process):

    def __init__(self, id, name, args=None, cwd=None):
        super(LocalProcess, self).__init__(id, name, args)
        self.cwd = cwd
        self.process = None
        self.input = None
        self.output = None
        self.error = None

    def execute(self):
        self.process = Popen(args=[self.name]+self.args, cwd=self.cwd, stdin=PIPE, stdout=PIPE, stderr=PIPE)
        self.input = self.process.stdin
        self.output = self.process.stdout
        self.error = self.process.stderr

    def wait(self):
        if self.process:
            return self.process.wait()
        return 0

    def kill(self):
        self.process.kill()

    def close_streams(self):
        self.input.close()
        self.output.close()
        self.error.close()
        pass

    def get_returncode(self):
        if self.process:
            return self.process.returncode
        return 0


class SSHRemoteProcess(LocalProcess):
    def __init__(self, id, name, args=None, cwd=None, forward_video=False, forward_audio=False,
                 compress_stream=False, target_host=None, target_port="22", ssh_user=''):
        self.forward_video = forward_video
        self.forward_audio = forward_audio
        self.compress_stream = compress_stream
        self.target_host = target_host
        self.target_port = target_port
        self.ssh_user = ssh_user

        self.ssh_args = self.get_ssh_args()
        args = self.ssh_args + [name] + args
        super(SSHRemoteProcess, self).__init__(id=id, name='/usr/bin/ssh', args=args, cwd=cwd)

    def get_ssh_args(self):
        args = []
        if self.forward_video:
            args.append('-Y')

        if self.compress_stream:
            args.append('-C')

        if not self.target_host:
            raise Exception('No target host for remote SSH process')

        if not self.ssh_user:
            raise Exception('No ssh user defined')

        return args + ['-f', '-p', self.target_port, '%s@%s' % (self.ssh_user, self.target_host)]


class RemoteProcess(Process):
    pass


# Process lock-files and IPC

class LockFile:

    def __init__(self, path, content=None, create=False, read_only=False):
        self.path = path
        self._content = content
        self.create_new = create
        self.file = None
        self.read_only = read_only
        self.__open()
        if create and content is not None:
            self.__write_content()

    def __write_content(self):
        self.file.seek(0)
        self.file.write(self._content)
        self.file.flush()

    def __open(self):
        if not self.exists():
            if not self.create_new:
                raise Exception('Lock file not found %s' % self.path)
            if not self.read_only:
                self.file = open(self.path, 'w')
        else:
            if self.create_new:
                raise Exception('Lock file already exists')
            self.file = open(self.path)

    def exists(self):
        return path.isfile(self.path)

    @property
    def content(self, value=None):
        if not self.file:
            raise Exception('Lock file not open')
        return self.file.read()

    @content.setter
    def content(self, value):
        if not self.file:
            raise Exception('Lock file not open')
        if self.read_only:
            raise Exception('Read-only file')
        self._content = value
        self.__write_content()


    def __del__(self):
        try:
            if self.file:
                self.file.close()
        except:
            pass

    def unlock(self):
        if self.path:
            remove(self.path)


class ProcessInfoFile(LockFile):

    def __init__(self, path, pid=None, create=False, with_info=None, read_only=False):
        self.pid = pid
        self.info = with_info or {}
        super(ProcessInfoFile, self).__init__(path=path, content=self.__generate_file_content(), create=create, read_only=read_only)

        if not create:
            self.__parse_file_content()

    def __generate_file_content(self):
        return '%d\n%s' % (self.pid or -1, json.dumps(self.info))

    def __parse_file_content(self):
        cnt = self.content.splitlines()
        self.pid = int(cnt[0])
        self.info = json.loads(cnt[1])

    def set_info(self, name, info):
        self.info[name] = info
        self.content = self.__generate_file_content()

    def get_info(self, name):
        return self.info.get(name)


def this_process_info_file(path, info=None, create=False):
    pid = getpid()
    try:
        return ProcessInfoFile(path=path, pid=pid, create=False)
    except Exception as e:
        if not create:
            raise e
        return ProcessInfoFile(path=path, pid=pid, create=True, with_info=info)


def open_process_lock_file(path, read_only=False):
    try:
        return ProcessInfoFile(path=path, create=False, read_only=read_only)
    except Exception as e:
        #logging.exception(e)
        return False



# Copyright 2016 Pavle Jonoski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import uuid4 as uuid4
import json

class Message:
    def __init__(self):
        self.id = None
        self.headers = {}
        self.data = None
    
    def __str__(self):
        return '<Message>(%s): Headers: %s; Data: %s;'%(self.id, self.headers, self.data)
    
    def __repr__(self):
        return self.__str__()

class MessageBuilder:
    def __init__(self):
        self.message = Message()
    
    def id(self, message_id):
        self.message.id = message_id
        return self
    
    def data(self, message_data):
        self.message.data = message_data
        return self
    
    def value(self, key, value):
        if not self.message.data:
            self.message.data = {}
        self.message.data[key] = value
        return self
    
    def header(self, name, value):
        self.message.headers[name] = value
        return self
    
    def build(self):
        if self.message.id is None:
            self.message.id = '%s'%uuid4()
        return self.message

def message(id=None, data=None):
    return MessageBuilder().id(id).data(data)


class DictEncoder(json.JSONEncoder):
    
    def default(self, o):
        try:
            return o.__dict__
        except:
            return None


def serialize(msg, indent=None):
    return json.dumps(msg, indent=indent, cls=DictEncoder)


def deserialize(smsg, as_type=None, strict=False):
    msg_type = as_type or Message
    dmsg = json.loads(smsg)
    
    msg = msg_type()
    
    for name, value in dmsg.items():
        if hasattr(msg, name):
            setattr(msg, name, value)
        elif strict:
            raise Exception('No such property %s' % name)
    
    return msg 


def deserialize_dict(dval, as_type, strict=None):
    val = as_type()
    
    for name, value in dval.items():
        if hasattr(val, name):
            setattr(val, name, value)
        elif strict:
            raise Exception('No such property %s' % name)
    
    return val




# Copyright 2016 Pavle Jonoski
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__author__ = 'pavle'


class App:

    def __init__(self, name=None, description=None, command=None, params=None, needs=None):
        self.name = name
        self.description = description
        self.command = command
        self.params = params or {}
        self.needs = needs or { "cpu": 0, "memory": 0, "network": 0, "disk": 0 }
